// Magloader 
//          by kumagai,masaaki

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAXSIZE 1280*1000 

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;

struct MagHeader
{
    byte header;
    byte machine;
    byte m_dependence;
    byte screen_mode;
    word sx;
    word sy;
    word ex;
    word ey;
    dword flaga;
    dword flagb;
    dword flagbsize;
    dword pixel;
    dword pixelsize;
};

inline word IW(byte *p)     // ƥ
{
    return ((word)p[0])+((word)p[1]*256);
};

inline dword IDW(byte *p)   // ƥ֥
{
    return 
      ((dword)p[0])+
	((dword)p[1])*256+
	  ((dword)p[2])*65536L+
	    ((dword)p[3])*16777216L;
};

// إå
MagHeader *BufferToHeader(byte *buffer)
{
    MagHeader *head=new MagHeader;
    
    head->header       = buffer[0];
    head->machine      = buffer[1];
    head->m_dependence = buffer[2];
    head->screen_mode  = buffer[3];
    head->sx           = IW(buffer+4);
    head->sy           = IW(buffer+6);
    head->ex           = IW(buffer+8);
    head->ey           = IW(buffer+10);
    head->flaga        = IDW(buffer+12);
    head->flagb        = IDW(buffer+16); 
    head->flagbsize    = IDW(buffer+20);
    head->pixel        = IDW(buffer+24); 
    head->pixelsize    = IDW(buffer+28);
    return head;
}
   
// MAG ǥɥơ֥
static int position[16][2]=
{ 
  {0,0},
  {1,0},
  {2,0},
  {4,0},
  {0,1},
  {1,1},
  {0,2},
  {1,2},
  {2,2},
  {0,4},
  {1,4},
  {2,4},
  {0,8},
  {1,8},
  {2,8},
  {0,16},
};

int pos_offset[16];

/* ---------------- Mag Loader --------------------*/

#define FBSize 800
char _comment[200];
char _user[20];

int MagLoad(byte *&source, byte *&palette,
	    char *filename,int &sizex,int &sizey,
	    char *&user, char *&comment,int &mode)
//unsigned short **source;      /*  Pixel Data          */
//unsigned char **palette;      /*  palette data grbgrb */
//char *filename;              /*  Output Filename     */   
//int *sizex;
//int *sizey;                   /*  Source Size:Pixel   */
//char **user;                  /*  User's name         */
//char **comment;               /*  Comment :: Sjis     */
//int *mode;                    /*  256:1 16:0          */
{
    int x,y,ycount,i;
    unsigned short *dest,*dest2;
    unsigned char *databuff,*filepoint;
    unsigned int filesize;
    
    int pixelsize,flagasize,flagbsize;
    int sx,sy,ex,ey;
    int faoff,fboff,poff;
    
    unsigned char *flaga,*flagb;
    unsigned short *pixel;
    unsigned char *_palette;
    
    unsigned char mask,flag;
    
    char realname[200];
    unsigned char flagbuf[FBSize];
    
    FILE *fp;
    
    
    /* ------------- file load section ------------------ */
    
    
    strcpy(realname,filename);
    if(strstr(realname,".mag")==NULL) 
      strcat(realname,".mag");
    
    fp=fopen(realname,"rb");
    if(fp==NULL)
      {
	  fprintf(stderr,"File cannot open '%s'.\n",realname);
	  return(-1);
      }
    
    fseek(fp,0,2);
    filesize=ftell(fp);
    fseek(fp,0,0);
    
    // եμ;
    databuff=(byte *)malloc(filesize+200);
    //                   ԡͽΰʣ
    fread(databuff,filesize,1,fp);
    fclose(fp);
    
    filepoint=databuff;
    
    // եޥå
    if(strncmp((char *)filepoint,"MAKI02  ",8)!=0)
      {
	  fprintf(stderr,"This file isn't mag file\n");
	  free(databuff);
	  return(-1);
      }
    filepoint+=8;
  
    // 桼̾
    strncpy(_user,(char *)filepoint,18);
    _user[18]='\0';
    user=_user;
    filepoint+=18;

    // ʬ
    {
	char *p=strchr((char *)filepoint,0x1a);
	if(p==NULL)
	  {
	      fprintf(stderr,"Broken Comment\n");
	      free(databuff);
	      return -1;
	  }
	*p='\0';
	comment=strdup((char *)filepoint);
	filepoint=(byte *)(p+1);
    }
    // Υ饤Ȥʤ
    {
	byte *_db;
	_db=(byte *)malloc(filesize);
	if(_db==NULL)
	  fprintf(stderr,"cannot alloc memory in alignment\n");
	memcpy(_db,filepoint,filesize);
	free(databuff);
	databuff=_db;
	filepoint=databuff;
    }


    // إå
    MagHeader *header=BufferToHeader(filepoint);
    

    // ѥå
    if((mode = (header->screen_mode >> 7))!=0)
      {
	  _palette=(byte *)malloc(768);
	  if(_palette==NULL)
	    {
		fprintf(stderr,"cannot allocate memory!");
		return(-1);
	    }
	  memcpy(_palette,filepoint+32,768);
      }
    else
      {
	  _palette=(byte *)malloc(48);
	  if(_palette==NULL)
	    {
		fprintf(stderr,"cannot allocate memory!");
		return(-1);
	    }
	  memcpy(_palette,filepoint+32,48);
      }
    
    palette = _palette;

    sx=header->sx;
    sy=header->sy;
    ex=header->ex;
    ey=header->ey;
    if((header->screen_mode & 0x80) != 0) 
      sizex=(ex-sx+1)/2;
    else
      sizex=(ex-sx+1)/4;

    sizey=(ey-sy+1);

    faoff=header->flaga;
    fboff=header->flagb;
    poff=header->pixel;

    for(i=0;i<16;i++)
      pos_offset[i]= position[i][0]+position[i][1]*sizex;

    flaga=filepoint+faoff;
    flagb=filepoint+fboff;
    (byte *)pixel=filepoint+poff;

    (byte *)dest=(byte *)malloc(sizex * sizey * 2);
    if(dest==NULL)
      {
	  fprintf(stderr,"Cannot alloc memory\n");
	  return(-1);
      }

    dest2=dest;
    source=(byte *)dest;

    for(i=0;i<FBSize;i++) flagbuf[i]=0;
    mask=0x80;
    
    for(y=0;y< sizey ;y++)
      {
	  for(x=0 ;x < sizex/2;x++)
	    {
		if((*flaga & mask) != 0)
		  {
		      flagbuf[x]=flagbuf[x]^(*flagb);
		      flagb++;
		  }
		mask = mask >> 1;
		if(mask==0)
		  {
		      mask=0x80;
		      flaga++;
		  }
		
		flag=flagbuf[x] >> 4;
		if(flag==0)
		  {
		      *dest2= *pixel;
		      pixel++;
		  }
		else 
		  {
		      *dest2= *(dest2-pos_offset[flag]);
		  }
		dest2++;
		
		flag=flagbuf[x] & 0xf;
		
		if(flag==0)
		  {
		      *dest2= *pixel;
		      pixel++;
		  }
		else 
		  {
		      *dest2= *(dest2-pos_offset[flag]);
		  }
		dest2++;
	    }
      }
  free(databuff);
  return(0);
}
